@@ -1,7 +1,7 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,25 +23,25 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * Ident (RFC1413) protocol support
- * $Id: ident.c,v 1.19 2003/08/06 22:03:32 castaglia Exp $
- */
+/* Ident (RFC1413) protocol support */
 
 #include "conf.h"
 
 static int ident_timeout;
 static pr_netio_stream_t *nstrm = NULL;
 
+static const char *trace_channel = "ident";
+
 static int ident_timeout_cb(CALLBACK_FRAME) {
   ident_timeout++;
 
-  if (nstrm)
+  if (nstrm) {
     /* Abort the NetIO stream, which will cause netio_poll (and thus
      * netio_read) to also abort.  This is similar to the way data connects
      * are aborted.
      */
     pr_netio_abort(nstrm);
+  }
 
   return 0;
 }
@@ -51,7 +51,7 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
   pool *tmp_pool = NULL;
   conn_t *ident_conn = NULL, *ident_io = NULL;
   char buf[256] = {'\0'}, *tok = NULL, *tmp = NULL;
-  int timer,i = 0;
+  int timerno, i = 0;
   int ident_port = pr_inet_getservport(p, "ident", "tcp");
 
   tmp_pool = make_sub_pool(p);
@@ -64,8 +64,9 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
   }
 
   /* Set up our timer before going any further. */
-  if ((timer = add_timer(PR_TUNABLE_TIMEOUTIDENT, -1, NULL,
-      (callback_t) ident_timeout_cb)) <= 0) {
+  timerno = pr_timer_add(PR_TUNABLE_TIMEOUTIDENT, -1, NULL,
+    (callback_t) ident_timeout_cb, "ident lookup");
+  if (timerno <= 0) {
     destroy_pool(tmp_pool);
     return pstrdup(p, ret);
   }
@@ -74,12 +75,18 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
     INPORT_ANY, FALSE);
   pr_inet_set_nonblock(tmp_pool, ident_conn);
 
-  if ((i = pr_inet_connect_nowait(tmp_pool, ident_conn, c->remote_addr,
-      ident_port)) < 0) {
-    remove_timer(timer, NULL);
+  i = pr_inet_connect_nowait(tmp_pool, ident_conn, c->remote_addr, ident_port);
+  if (i < 0) {
+    int xerrno = errno;
+
+    pr_timer_remove(timerno, ANY_MODULE);
     pr_inet_close(tmp_pool, ident_conn);
-    log_debug(DEBUG6, "ident connection failed: %s", strerror(errno));
+    pr_trace_msg(trace_channel, 5, "connection to %s, port %d failed: %s",
+      pr_netaddr_get_ipstr(c->remote_addr), ident_port, strerror(xerrno));
+
     destroy_pool(tmp_pool);
+
+    errno = xerrno;
     return pstrdup(p, ret);
   }
 
@@ -92,48 +99,72 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
     switch (pr_netio_poll(nstrm)) {
 
       /* Aborted, timed out */
-      case 1:
+      case 1: {
         if (ident_timeout) {
-          remove_timer(timer, NULL);
+          pr_timer_remove(timerno, ANY_MODULE);
           pr_netio_close(nstrm);
           pr_inet_close(tmp_pool, ident_conn);
-          log_debug(DEBUG6, "ident lookup timed out, returning '%s'", ret);
+
+          pr_trace_msg(trace_channel, 5, "lookup timed out, returning '%s'",
+            ret);
           destroy_pool(tmp_pool);
           return pstrdup(p, ret);
         }
         break;
+      }
 
       /* Error. */
-      case -1:
-        remove_timer(timer, NULL);
+      case -1: {
+        int xerrno = errno;
+
+        pr_timer_remove(timerno, ANY_MODULE);
         pr_netio_close(nstrm);
         pr_inet_close(tmp_pool, ident_conn);
-        log_debug(DEBUG6, "ident lookup failed (%s), returning '%s'",
-          strerror(errno), ret);
+
+        pr_trace_msg(trace_channel, 6, "lookup failed (%s), returning '%s'",
+          strerror(xerrno), ret);
         destroy_pool(tmp_pool);
+
+        errno = xerrno;
         return pstrdup(p, ret);
+      }
 
       /* Connected. */
-      default:
+      default: {
         ident_conn->mode = CM_OPEN;
 
         if (pr_inet_get_conn_info(ident_conn, ident_conn->listen_fd) < 0) {
-          remove_timer(timer, NULL);
+          int xerrno = errno;
+
+          pr_timer_remove(timerno, ANY_MODULE);
           pr_netio_close(nstrm);
           pr_inet_close(tmp_pool, ident_conn);
-          log_debug(DEBUG2, "ident lookup timed out, returning '%s'", ret);
+
+          pr_trace_msg(trace_channel, 2,
+            "lookup timed out (%s), returning '%s'", strerror(xerrno), ret);
           destroy_pool(tmp_pool);
+
+          errno = xerrno;
           return pstrdup(p, ret);
         }
         break;
+      }
     }
   }
 
-  if ((ident_io = pr_inet_openrw(tmp_pool, ident_conn, NULL, PR_NETIO_STRM_OTHR,
-      -1, -1, -1, FALSE)) == NULL) {
-    remove_timer(timer, NULL);
+  ident_io = pr_inet_openrw(tmp_pool, ident_conn, NULL, PR_NETIO_STRM_OTHR,
+    -1, -1, -1, FALSE);
+  if (ident_io == NULL) {
+    int xerrno = errno;
+
+    pr_timer_remove(timerno, ANY_MODULE);
     pr_inet_close(tmp_pool, ident_conn);
+
+    pr_trace_msg(trace_channel, 3, "failed opening read/write connection: %s",
+      strerror(xerrno));
     destroy_pool(tmp_pool);
+
+    errno = xerrno;
     return pstrdup(p, ret);
   }
 
@@ -153,20 +184,29 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
    * automatically restart syscalls after the SIGALRM signal.
    */
 
+  pr_trace_msg(trace_channel, 4, "reading response from remote ident server");
+
   if (pr_netio_gets(buf, sizeof(buf), ident_io->instrm)) {
     strip_end(buf, "\r\n");
 
+    pr_trace_msg(trace_channel, 6, "received '%s' from remote ident server",
+      buf);
+
     tmp = buf;
     tok = get_token(&tmp, ":");
     if (tok && (tok = get_token(&tmp, ":"))) {
-      while (*tok && isspace((int) *tok))
+      while (*tok && isspace((int) *tok)) {
+        pr_signals_handle();
         tok++;
+      }
       strip_end(tok, " \t");
 
       if (strcasecmp(tok, "ERROR") == 0) {
         if (tmp) {
-          while (*tmp && isspace((int) *tmp))
+          while (*tmp && isspace((int) *tmp)) {
+            pr_signals_handle();
             tmp++;
+          }
 	  strip_end(tmp, " \t");
           if (strcasecmp(tmp, "HIDDEN-USER") == 0)
             ret = "HIDDEN-USER";
@@ -175,8 +215,10 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
       } else if (strcasecmp(tok, "USERID") == 0) {
         if (tmp && (tok = get_token(&tmp, ":"))) {
           if (tmp) {
-            while (*tmp && isspace((int) *tmp))
+            while (*tmp && isspace((int) *tmp)) {
+              pr_signals_handle();
               tmp++;
+            }
             strip_end(tmp, " \t");
             ret = tmp;
           }
@@ -185,7 +227,7 @@ char *pr_ident_lookup(pool *p, conn_t *c) {
     }
   }
 
-  remove_timer(timer, NULL);
+  pr_timer_remove(timerno, ANY_MODULE);
   pr_inet_close(tmp_pool, ident_io);
   pr_inet_close(tmp_pool, ident_conn);
   destroy_pool(tmp_pool);
