@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2016 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -24,10 +24,7 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * Resource allocation code
- * $Id: pool.c,v 1.42 2004/01/09 04:41:59 castaglia Exp $
- */
+/* Resource allocation code */
 
 #include "conf.h"
 
@@ -46,99 +43,120 @@ union align {
 union block_hdr {
   union align a;
 
-  /* Actual header */
+  /* Padding */
+#if defined(_LP64) || defined(__LP64__)
+  char pad[32];
+#endif
 
+  /* Actual header */
   struct {
-    char *endp;
+    void *endp;
     union block_hdr *next;
-    char *first_avail;
+    void *first_avail;
   } h;
 };
 
-union block_hdr *block_freelist = NULL;
+static union block_hdr *block_freelist = NULL;
 
 /* Statistics */
 static unsigned int stat_malloc = 0;	/* incr when malloc required */
 static unsigned int stat_freehit = 0;	/* incr when freelist used */
 
+#ifdef PR_USE_DEVEL
+static const char *trace_channel = "pool";
+#endif /* PR_USE_DEVEL */
+
+#ifdef PR_USE_DEVEL
+/* Debug flags */
+static int debug_flags = 0;
+
+static void oom_printf(const char *fmt, ...) {
+  char buf[PR_TUNABLE_BUFFER_SIZE];
+  va_list msg;
+
+  memset(buf, '\0', sizeof(buf));
+
+  va_start(msg, fmt);
+  vsnprintf(buf, sizeof(buf), fmt, msg);
+  va_end(msg);
+
+  buf[sizeof(buf)-1] = '\0';
+  fprintf(stderr, "%s\n", buf);
+}
+#endif /* PR_USE_DEVEL */
+
 /* Lowest level memory allocation functions
  */
 
-static void *null_alloc(size_t size) {
-  void *ret = 0;
-
-  if (size == 0)
-    ret = malloc(size);
-  if (ret == 0) {
-    pr_log_pri(PR_LOG_ERR, "fatal: Memory exhausted");
-    exit(1);
+static void null_alloc(void) {
+  pr_log_pri(PR_LOG_ALERT, "Out of memory!");
+#ifdef PR_USE_DEVEL
+  if (debug_flags & PR_POOL_DEBUG_FL_OOM_DUMP_POOLS) {
+    pr_pool_debug_memory(oom_printf);
   }
+#endif
 
-  return ret;
+  exit(1);
 }
 
 static void *smalloc(size_t size) {
-  void *ret;
+  void *res;
 
-  ret = malloc(size);
-  if (ret == 0)
-    ret = null_alloc(size);
+  if (size == 0) {
+    /* Avoid zero-length malloc(); on non-POSIX systems, the behavior is
+     * not dependable.  And on POSIX systems, malloc(3) might still return
+     * a "unique pointer" for a zero-length allocation (or NULL).
+     *
+     * Either way, a zero-length allocation request here means that someone
+     * is doing something they should not be doing.
+     */
+    null_alloc();
+  }
 
-  return ret;
+  res = malloc(size);
+  if (res == NULL) {
+    null_alloc();
+  }
+
+  return res;
 }
 
-#if 0
-void *scalloc(size_t num, size_t size) {
-  void *ret;
-
-  ret = calloc(num,size);
-  if (ret == 0)
-    ret = null_alloc(num * size);
-  return ret;
-}
-
-void *srealloc(void *p, size_t size) {
-  if (p == 0)
-    return smalloc(size);
-  p = realloc(p,size);
-  if (p == 0)
-    p = null_alloc(size);
-  return p;
-}
-#endif
-
 /* Grab a completely new block from the system pool.  Relies on malloc()
  * to return truly aligned memory.
  */
-
-static union block_hdr *malloc_block(int size) {
+static union block_hdr *malloc_block(size_t size) {
   union block_hdr *blok =
     (union block_hdr *) smalloc(size + sizeof(union block_hdr));
 
   blok->h.next = NULL;
   blok->h.first_avail = (char *) (blok + 1);
-  blok->h.endp = size + blok->h.first_avail;
+  blok->h.endp = size + (char *) blok->h.first_avail;
 
   return blok;
 }
 
-static void chk_on_blk_list(union block_hdr *blok, union block_hdr *free_blk) {
+static void chk_on_blk_list(union block_hdr *blok, union block_hdr *free_blk,
+    const char *pool_tag) {
+
+#ifdef PR_USE_DEVEL
   /* Debug code */
 
   while (free_blk) {
-    if (free_blk == blok) {
-      pr_log_pri(PR_LOG_ERR, "Fatal: DEBUG: Attempt to free already free block "
-       "in chk_on_blk_list()");
-      exit(1);
+    if (free_blk != blok) {
+      free_blk = free_blk->h.next;
+      continue;
     }
 
-    free_blk = free_blk->h.next;
+    pr_log_pri(PR_LOG_WARNING, "fatal: DEBUG: Attempt to free already free "
+     "block in pool '%s'", pool_tag ? pool_tag : "<unnamed>");
+    exit(1);
   }
+#endif /* PR_USE_DEVEL */
 }
 
 /* Free a chain of blocks -- _must_ call with alarms blocked. */
 
-static void free_blocks(union block_hdr *blok) {
+static void free_blocks(union block_hdr *blok, const char *pool_tag) {
   /* Puts new blocks at head of block list, point next pointer of
    * last block in chain to free blocks we already had.
    */
@@ -153,12 +171,12 @@ static void free_blocks(union block_hdr *blok) {
   /* Adjust first_avail pointers */
 
   while (blok->h.next) {
-    chk_on_blk_list(blok, old_free_list);
+    chk_on_blk_list(blok, old_free_list, pool_tag);
     blok->h.first_avail = (char *) (blok + 1);
     blok = blok->h.next;
   }
 
-  chk_on_blk_list(blok, old_free_list);
+  chk_on_blk_list(blok, old_free_list, pool_tag);
   blok->h.first_avail = (char *) (blok + 1);
   blok->h.next = old_free_list;
 }
@@ -184,17 +202,16 @@ static union block_hdr *new_block(int minsz, int exact) {
   /* Check if we have anything of the requested size on our free list first...
    */
   while (blok) {
-    if (minsz <= blok->h.endp - blok->h.first_avail) {
+    if (minsz <= ((char *) blok->h.endp - (char *) blok->h.first_avail)) {
       *lastptr = blok->h.next;
       blok->h.next = NULL;
 
       stat_freehit++;
       return blok;
-
-    } else {
-      lastptr = &blok->h.next;
-      blok = blok->h.next;
     }
+
+    lastptr = &blok->h.next;
+    blok = blok->h.next;
   }
 
   /* Nope...damn.  Have to malloc() a new one. */
@@ -202,33 +219,20 @@ static union block_hdr *new_block(int minsz, int exact) {
   return malloc_block(minsz);
 }
 
-/* Accounting */
-
-static unsigned long bytes_in_block_list(union block_hdr *blok) {
-  unsigned long size = 0;
-
-  while (blok) {
-    size += blok->h.endp - (char *) (blok + 1);
-    blok = blok->h.next;
-  }
-
-  return size;
-}
-
 struct cleanup;
 
 static void run_cleanups(struct cleanup *);
 
 /* Pool internal and management */
 
-struct pool {
+struct pool_rec {
   union block_hdr *first;
   union block_hdr *last;
   struct cleanup *cleanups;
-  struct pool *sub_pools;
-  struct pool *sub_next;
-  struct pool *sub_prev;
-  struct pool *parent;
+  struct pool_rec *sub_pools;
+  struct pool_rec *sub_next;
+  struct pool_rec *sub_prev;
+  struct pool_rec *parent;
   char *free_first_avail;
   const char *tag;
 };
@@ -241,89 +245,174 @@ pool *global_config_pool = NULL;
  * aligned).
  */
 
-#define POOL_HDR_CLICKS (1 + ((sizeof(struct pool) - 1) / CLICK_SZ))
+#define POOL_HDR_CLICKS (1 + ((sizeof(struct pool_rec) - 1) / CLICK_SZ))
 #define POOL_HDR_BYTES (POOL_HDR_CLICKS * CLICK_SZ)
 
-/* walk all pools, starting with top level permanent pool, displaying a
+#ifdef PR_USE_DEVEL
+
+static unsigned long blocks_in_block_list(union block_hdr *blok) {
+  unsigned long count = 0;
+
+  while (blok) {
+    count++;
+    blok = blok->h.next;
+  }
+
+  return count;
+}
+
+static unsigned long bytes_in_block_list(union block_hdr *blok) {
+  unsigned long size = 0;
+
+  while (blok) {
+    size += ((char *) blok->h.endp - (char *) (blok + 1));
+    blok = blok->h.next;
+  }
+
+  return size;
+}
+
+static unsigned int subpools_in_pool(pool *p) {
+  unsigned int count = 0;
+  pool *iter;
+
+  if (p->sub_pools == NULL)
+    return 0;
+
+  for (iter = p->sub_pools; iter; iter = iter->sub_next) {
+    /* Count one for the current subpool (iter). */
+    count += (subpools_in_pool(iter) + 1);
+  }
+
+  return count;
+}
+
+/* Walk all pools, starting with top level permanent pool, displaying a
  * tree.
  */
-
-static long __walk_pools(pool *p, int level,
+static long walk_pools(pool *p, unsigned long level,
     void (*debugf)(const char *, ...)) {
   char _levelpad[80] = "";
   long total = 0;
 
-  if (!p)
+  if (p == NULL) {
     return 0;
+  }
 
   if (level > 1) {
     memset(_levelpad, ' ', sizeof(_levelpad)-1);
-    if ((level - 1) * 3 >= sizeof(_levelpad))
+
+    if ((level - 1) * 3 >= sizeof(_levelpad)) {
       _levelpad[sizeof(_levelpad)-1] = 0;
-    else
+
+    } else {
       _levelpad[(level - 1) * 3] = '\0';
+    }
   }
 
+  /* The emitted message is:
+   *
+   *  <pool-tag> [pool-ptr] (n B, m L, r P)
+   *
+   * where n is the number of bytes (B), m is the number of allocated blocks
+   * in the pool list (L), and r is the number of sub-pools (P).
+   */
+
   for (; p; p = p->sub_next) {
     total += bytes_in_block_list(p->first);
-    if (level == 0)
-      debugf("%s (%lu bytes)", p->tag ? p->tag : "[none]",
-        bytes_in_block_list(p->first));
+    if (level == 0) {
+      debugf("%s [%p] (%lu B, %lu L, %u P)",
+        p->tag ? p->tag : "<unnamed>", p, bytes_in_block_list(p->first),
+        blocks_in_block_list(p->first), subpools_in_pool(p));
 
-    else
-      debugf("%s\\- %s (%lu bytes)", _levelpad,
-        p->tag ? p->tag : "[none]", bytes_in_block_list(p->first));
+    } else {
+      debugf("%s + %s [%p] (%lu B, %lu L, %u P)", _levelpad,
+        p->tag ? p->tag : "<unnamed>", p, bytes_in_block_list(p->first),
+        blocks_in_block_list(p->first), subpools_in_pool(p));
+    }
 
     /* Recurse */
-    if (p->sub_pools)
-      total += __walk_pools(p->sub_pools, level+1, debugf);
+    if (p->sub_pools) {
+      total += walk_pools(p->sub_pools, level+1, debugf);
+    }
   }
 
   return total;
 }
 
 static void debug_pool_info(void (*debugf)(const char *, ...)) {
-  if (block_freelist)
+  if (block_freelist) {
     debugf("Free block list: %lu bytes",
       bytes_in_block_list(block_freelist));
-  else
-    debugf("Free block list: EMPTY");
 
-  debugf("%u count blocks allocated", stat_malloc);
-  debugf("%u count blocks reused", stat_freehit);
+  } else {
+    debugf("Free block list: empty");
+  }
+
+  debugf("%u blocks allocated", stat_malloc);
+  debugf("%u blocks reused", stat_freehit);
+}
+
+static void pool_printf(const char *fmt, ...) {
+  char buf[PR_TUNABLE_BUFFER_SIZE];
+  va_list msg;
+
+  memset(buf, '\0', sizeof(buf));
+
+  va_start(msg, fmt);
+  vsnprintf(buf, sizeof(buf), fmt, msg);
+  va_end(msg);
+
+  buf[sizeof(buf)-1] = '\0';
+  pr_trace_msg(trace_channel, 5, "%s", buf);
 }
 
 void pr_pool_debug_memory(void (*debugf)(const char *, ...)) {
+  if (debugf == NULL) {
+    debugf = pool_printf;
+  }
+
   debugf("Memory pool allocation:");
-  debugf("Total %lu bytes allocated",
-    __walk_pools(permanent_pool, 0, debugf));
+  debugf("Total %lu bytes allocated", walk_pools(permanent_pool, 0, debugf));
   debug_pool_info(debugf);
 }
 
+int pr_pool_debug_set_flags(int flags) {
+  if (flags < 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  debug_flags = flags;
+  return 0;
+}
+#endif /* PR_USE_DEVEL */
+
 void pr_pool_tag(pool *p, const char *tag) {
-  if (!p || !tag)
+  if (p == NULL ||
+      tag == NULL) {
     return;
+  }
 
   p->tag = tag;
 }
 
 /* Release the entire free block list */
 static void pool_release_free_block_list(void) {
-  union block_hdr *blok,*next;
+  union block_hdr *blok = NULL, *next = NULL;
 
   pr_alarms_block();
 
-  blok = block_freelist;
-  if (blok) {
-    for (next = blok->h.next; next; blok = next, next = blok->h.next)
-      free(blok);
+  for (blok = block_freelist; blok; blok = next) {
+    next = blok->h.next;
+    free(blok);
   }
   block_freelist = NULL;
 
   pr_alarms_unblock();
 }
 
-struct pool *make_sub_pool(struct pool *p) {
+struct pool_rec *make_sub_pool(struct pool_rec *p) {
   union block_hdr *blok;
   pool *new_pool;
 
@@ -332,9 +421,9 @@ struct pool *make_sub_pool(struct pool *p) {
   blok = new_block(0, FALSE);
 
   new_pool = (pool *) blok->h.first_avail;
-  blok->h.first_avail += POOL_HDR_BYTES;
+  blok->h.first_avail = POOL_HDR_BYTES + (char *) blok->h.first_avail;
 
-  memset(new_pool, 0, sizeof(struct pool));
+  memset(new_pool, 0, sizeof(struct pool_rec));
   new_pool->free_first_avail = blok->h.first_avail;
   new_pool->first = new_pool->last = blok;
 
@@ -353,18 +442,18 @@ struct pool *make_sub_pool(struct pool *p) {
   return new_pool;
 }
 
-struct pool *pr_pool_create_sz(struct pool *p, int sz) {
+struct pool_rec *pr_pool_create_sz(struct pool_rec *p, size_t sz) {
   union block_hdr *blok;
   pool *new_pool;
 
   pr_alarms_block();
 
-  blok = new_block(sz, TRUE);
+  blok = new_block(sz + POOL_HDR_BYTES, TRUE);
 
   new_pool = (pool *) blok->h.first_avail;
-  blok->h.first_avail += POOL_HDR_BYTES;
+  blok->h.first_avail = POOL_HDR_BYTES + (char *) blok->h.first_avail;
 
-  memset(new_pool, 0, sizeof(struct pool));
+  memset(new_pool, 0, sizeof(struct pool_rec));
   new_pool->free_first_avail = blok->h.first_avail;
   new_pool->first = new_pool->last = blok;
 
@@ -386,8 +475,10 @@ struct pool *pr_pool_create_sz(struct pool *p, int sz) {
 /* Initialize the pool system by creating the base permanent_pool. */
 
 void init_pools(void) {
-  if (!permanent_pool)
+  if (permanent_pool == NULL) {
     permanent_pool = make_sub_pool(NULL);
+  }
+
   pr_pool_tag(permanent_pool, "permanent_pool");
 }
 
@@ -397,11 +488,12 @@ void free_pools(void) {
   pool_release_free_block_list();
 }
 
-static void clear_pool(struct pool *p) {
+static void clear_pool(struct pool_rec *p) {
 
   /* Sanity check. */
-  if (!p)
+  if (p == NULL) {
     return;
+  }
 
   pr_alarms_block();
 
@@ -410,11 +502,13 @@ static void clear_pool(struct pool *p) {
   p->cleanups = NULL;
 
   /* Destroy subpools. */
-  while (p->sub_pools)
+  while (p->sub_pools) {
     destroy_pool(p->sub_pools);
+  }
+
   p->sub_pools = NULL;
 
-  free_blocks(p->first->h.next);
+  free_blocks(p->first->h.next, p->tag);
   p->first->h.next = NULL;
 
   p->last = p->first;
@@ -424,61 +518,75 @@ static void clear_pool(struct pool *p) {
 }
 
 void destroy_pool(pool *p) {
-  if (p == NULL)
+  if (p == NULL) {
     return;
+  }
 
   pr_alarms_block();
 
   if (p->parent) {
-    if (p->parent->sub_pools == p)
+    if (p->parent->sub_pools == p) {
       p->parent->sub_pools = p->sub_next;
+    }
 
-    if (p->sub_prev)
+    if (p->sub_prev) {
       p->sub_prev->sub_next = p->sub_next;
+    }
 
-    if (p->sub_next)
+    if (p->sub_next) {
       p->sub_next->sub_prev = p->sub_prev;
+    }
   }
+
   clear_pool(p);
-  free_blocks(p->first);
+  free_blocks(p->first, p->tag);
 
   pr_alarms_unblock();
-}
 
-#if 0
-/* NOTE: not used at the moment */
-static long bytes_in_pool(pool *p) {
-  return bytes_in_block_list(p->first);
+#ifdef PR_DEVEL_NO_POOL_FREELIST
+  /* If configured explicitly to do so, call free(3) on the freelist after
+   * a pool is destroyed.  This can be useful for tracking down use-after-free
+   * and other memory issues using libraries such as dmalloc.
+   */
+  pool_release_free_block_list();
+#endif /* PR_EVEL_NO_POOL_FREELIST */
 }
 
-static long bytes_in_free_blocks(void) {
-  return bytes_in_block_list(block_freelist);
-}
-#endif
-
 /* Allocation interface...
  */
 
-static void *alloc_pool(struct pool *p, int reqsz, int exact) {
-
+static void *alloc_pool(struct pool_rec *p, size_t reqsz, int exact) {
   /* Round up requested size to an even number of aligned units */
-  int nclicks = 1 + ((reqsz - 1) / CLICK_SZ);
-  int sz = nclicks * CLICK_SZ;
+  size_t nclicks = 1 + ((reqsz - 1) / CLICK_SZ);
+  size_t sz = nclicks * CLICK_SZ;
+  union block_hdr *blok;
+  char *first_avail, *new_first_avail;
 
   /* For performance, see if space is available in the most recently
    * allocated block.
    */
 
-  union block_hdr *blok = p->last;
-  char *first_avail = blok->h.first_avail;
-  char *new_first_avail;
-
-  if (reqsz <= 0)
+  blok = p->last;
+  if (blok == NULL) {
+    errno = EINVAL;
     return NULL;
+  }
+
+  first_avail = blok->h.first_avail;
+
+  if (reqsz == 0) {
+    /* Don't try to allocate memory of zero length.
+     *
+     * This should NOT happen normally; if it does, by returning NULL we
+     * almost guarantee a null pointer dereference.
+     */
+    errno = EINVAL;
+    return NULL;
+  }
 
   new_first_avail = first_avail + sz;
 
-  if (new_first_avail <= blok->h.endp) {
+  if (new_first_avail <= (char *) blok->h.endp) {
     blok->h.first_avail = new_first_avail;
     return (void *) first_avail;
   }
@@ -491,137 +599,53 @@ static void *alloc_pool(struct pool *p, int reqsz, int exact) {
   p->last = blok;
 
   first_avail = blok->h.first_avail;
-  blok->h.first_avail += sz;
+  blok->h.first_avail = sz + (char *) blok->h.first_avail;
 
   pr_alarms_unblock();
   return (void *) first_avail;
 }
 
-void *palloc(struct pool *p, int sz) {
+void *palloc(struct pool_rec *p, size_t sz) {
   return alloc_pool(p, sz, FALSE);
 }
 
-void *pallocsz(struct pool *p, int sz) {
+void *pallocsz(struct pool_rec *p, size_t sz) {
   return alloc_pool(p, sz, TRUE);
 }
 
-void *pcalloc(struct pool *p, int sz) {
-  void *res = palloc(p, sz);
+void *pcalloc(struct pool_rec *p, size_t sz) {
+  void *res;
+
+  res = palloc(p, sz);
   memset(res, '\0', sz);
+
   return res;
 }
 
-void *pcallocsz(struct pool *p, int sz) {
-  void *res = pallocsz(p, sz);
+void *pcallocsz(struct pool_rec *p, size_t sz) {
+  void *res;
+
+  res = pallocsz(p, sz);
   memset(res, '\0', sz);
+
   return res;
 }
 
-char *pstrdup(struct pool *p, const char *s) {
-  char *res;
-  size_t len;
+/* Array functions */
 
-  if (!s)
+array_header *make_array(pool *p, unsigned int nelts, size_t elt_size) {
+  array_header *res;
+
+  if (p == NULL ||
+      elt_size == 0) {
+    errno = EINVAL;
     return NULL;
-
-  len = strlen(s) + 1;
-
-  res = palloc(p, len);
-  sstrncpy(res, s, len);
-  return res;
-}
-
-char *pstrndup(struct pool *p, const char *s, int n) {
-  char *res;
-
-  if (!s)
-    return NULL;
-
-  res = palloc(p, n + 1);
-  sstrncpy(res, s, n + 1);
-  return res;
-}
-
-char *pdircat(pool *p, ...) {
-  char *argp, *res;
-  char last;
-
-  int len = 0, count = 0;
-  va_list dummy;
-
-  va_start(dummy, p);
-
-  last = 0;
-
-  while ((res = va_arg(dummy, char *)) != NULL) {
-    /* If the first argument is "", we have to account for a leading /
-     * which must be added.
-     */
-    if (!count++ && !*res)
-      len++;
-    else if (last && last != '/' && *res != '/')
-      len++;
-    else if (last && last == '/' && *res == '/')
-      len--;
-    len += strlen(res);
-    last = (*res ? res[strlen(res) - 1] : 0);
   }
 
-  va_end(dummy);
-  res = (char *) pcalloc(p, len + 1);
+  res = palloc(p, sizeof(array_header));
 
-  va_start(dummy, p);
-
-  last = 0;
-
-  while ((argp = va_arg(dummy, char *)) != NULL) {
-    if (last && last == '/' && *argp == '/')
-      argp++;
-    else if (last && last != '/' && *argp != '/')
-      sstrcat(res, "/", len + 1);
-
-    sstrcat(res, argp, len + 1);
-    last = (*res ? res[strlen(res) - 1] : 0);
-  }
-
-  va_end(dummy);
-
-  return res;
-}
-
-char *pstrcat(pool *p, ...) {
-  char *argp, *res;
-
-  size_t len = 0;
-  va_list dummy;
-
-  va_start(dummy, p);
-
-  while ((res = va_arg(dummy, char *)) != NULL)
-    len += strlen(res);
-
-  va_end(dummy);
-
-  res = (char *) pcalloc(p, len + 1);
-
-  va_start(dummy, p);
-
-  while ((argp = va_arg(dummy, char *)) != NULL)
-    sstrcat(res, argp, len + 1);
-
-  va_end(dummy);
-
-  return res;
-}
-
-/*
- * Array functions
- */
-
-array_header *make_array(pool *p, int nelts, int elt_size) {
-  array_header *res = (array_header *) palloc(p, sizeof(array_header));
-
-  if (nelts < 1) nelts = 1;
+  if (nelts < 1)
+    nelts = 1;
 
   res->elts = pcalloc(p, nelts * elt_size);
   res->pool = p;
@@ -632,7 +656,21 @@ array_header *make_array(pool *p, int nelts, int elt_size) {
   return res;
 }
 
+void clear_array(array_header *arr) {
+  if (arr == NULL) {
+    return;
+  }
+
+  arr->elts = pcalloc(arr->pool, arr->nalloc * arr->elt_size);
+  arr->nelts = 0;
+}
+
 void *push_array(array_header *arr) {
+  if (arr == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   if (arr->nelts == arr->nalloc) {
     char *new_data = pcalloc(arr->pool, arr->nalloc * arr->elt_size * 2);
 
@@ -642,21 +680,32 @@ void *push_array(array_header *arr) {
   }
 
   ++arr->nelts;
-  return ((char *)arr->elts) + (arr->elt_size * (arr->nelts - 1));
+  return ((char *) arr->elts) + (arr->elt_size * (arr->nelts - 1));
 }
 
-void array_cat(array_header *dst, const array_header *src)
-{
-  int elt_size = dst->elt_size;
+int array_cat2(array_header *dst, const array_header *src) {
+  size_t elt_size;
+
+  if (dst == NULL ||
+      src == NULL) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  elt_size = dst->elt_size;
 
   if (dst->nelts + src->nelts > dst->nalloc) {
-    int new_size = dst->nalloc * 2;
+    size_t new_size;
     char *new_data;
 
-    if (new_size == 0) ++new_size;
+    new_size = dst->nalloc * 2;
+    if (new_size == 0) {
+      ++new_size;
+    }
 
-    while (dst->nelts + src->nelts > new_size)
+    while ((dst->nelts + src->nelts) > new_size) {
       new_size *= 2;
+    }
 
     new_data = pcalloc(dst->pool, elt_size * new_size);
     memcpy(new_data, dst->elts, dst->nalloc * elt_size);
@@ -665,14 +714,27 @@ void array_cat(array_header *dst, const array_header *src)
     dst->nalloc = new_size;
   }
 
-  memcpy(((char *)dst->elts) + dst->nelts * elt_size, (char *)src->elts,
+  memcpy(((char *) dst->elts) + (dst->nelts * elt_size), (char *) src->elts,
          elt_size * src->nelts);
   dst->nelts += src->nelts;
+
+  return 0;
 }
 
-array_header *copy_array(pool *p, const array_header *arr)
-{
-  array_header *res = make_array(p,arr->nalloc,arr->elt_size);
+void array_cat(array_header *dst, const array_header *src) {
+  (void) array_cat2(dst, src);
+}
+
+array_header *copy_array(pool *p, const array_header *arr) {
+  array_header *res;
+
+  if (p == NULL ||
+      arr == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  res = make_array(p, arr->nalloc, arr->elt_size);
 
   memcpy(res->elts, arr->elts, arr->elt_size * arr->nelts);
   res->nelts = arr->nelts;
@@ -680,20 +742,34 @@ array_header *copy_array(pool *p, const array_header *arr)
 }
 
 /* copy an array that is assumed to consist solely of strings */
-array_header *copy_array_str(pool *p, const array_header *arr)
-{
-  array_header *res = copy_array(p,arr);
-  int i;
+array_header *copy_array_str(pool *p, const array_header *arr) {
+  register unsigned int i;
+  array_header *res;
+
+  if (p == NULL ||
+      arr == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  res = copy_array(p, arr);
 
   for (i = 0; i < arr->nelts; i++)
-    ((char **)res->elts)[i] = pstrdup(p, ((char **)res->elts)[i]);
+    ((char **) res->elts)[i] = pstrdup(p, ((char **) res->elts)[i]);
 
   return res;
 }
 
-array_header *copy_array_hdr(pool *p, const array_header *arr)
-{
-  array_header *res = (array_header *)palloc(p,sizeof(array_header));
+array_header *copy_array_hdr(pool *p, const array_header *arr) {
+  array_header *res;
+
+  if (p == NULL ||
+      arr == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  res = palloc(p, sizeof(array_header));
 
   res->elts = arr->elts;
   res->pool = p;
@@ -704,30 +780,42 @@ array_header *copy_array_hdr(pool *p, const array_header *arr)
   return res;
 }
 
-array_header *append_arrays(pool *p,
-                            const array_header *first,
-			    const array_header *second)
-{
-  array_header *res = copy_array_hdr(p,first);
+array_header *append_arrays(pool *p, const array_header *first,
+    const array_header *second) {
+  array_header *res;
 
-  array_cat(res,second);
+  if (p == NULL ||
+      first == NULL ||
+      second == NULL) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  res = copy_array_hdr(p, first);
+
+  array_cat(res, second);
   return res;
 }
 
-/*
- * Generic cleanups
- */
+/* Generic cleanups */
 
 typedef struct cleanup {
   void *data;
   void (*plain_cleanup_cb)(void *);
   void (*child_cleanup_cb)(void *);
   struct cleanup *next;
+
 } cleanup_t;
 
 void register_cleanup(pool *p, void *data, void (*plain_cleanup_cb)(void*),
     void (*child_cleanup_cb)(void *)) {
-  cleanup_t *c = pcalloc(p, sizeof(cleanup_t));
+  cleanup_t *c;
+
+  if (p == NULL) {
+    return;
+  }
+
+  c = pcalloc(p, sizeof(cleanup_t));
   c->data = data;
   c->plain_cleanup_cb = plain_cleanup_cb;
   c->child_cleanup_cb = child_cleanup_cb;
@@ -738,11 +826,18 @@ void register_cleanup(pool *p, void *data, void (*plain_cleanup_cb)(void*),
 }
 
 void unregister_cleanup(pool *p, void *data, void (*cleanup_cb)(void *)) {
-  cleanup_t *c = p->cleanups;
-  cleanup_t **lastp = &p->cleanups;
+  cleanup_t *c, **lastp;
+
+  if (p == NULL) {
+    return;
+  }
+
+  c = p->cleanups;
+  lastp = &p->cleanups;
 
   while (c) {
-    if (c->data == data && c->plain_cleanup_cb == cleanup_cb) {
+    if (c->data == data &&
+        (c->plain_cleanup_cb == cleanup_cb || cleanup_cb == NULL)) {
 
       /* Remove the given cleanup by pointing the previous next pointer to
        * the matching cleanup's next pointer.
@@ -756,142 +851,12 @@ void unregister_cleanup(pool *p, void *data, void (*cleanup_cb)(void *)) {
   }
 }
 
-/* NOTE: unused. */
-#if 0
-void run_cleanup(pool *p, void *data, void (*cleanup_cb)(void *)) {
-  pr_alarms_block();
-
-  /* Run the given cleanup callback. */
-  (*cleanup_cb)(data);
-
-  /* Remove it. */
-  unregister_cleanup(p, data, cleanup_cb);
-
-  pr_alarms_unblock();
-}
-#endif
-
 static void run_cleanups(cleanup_t *c) {
   while (c) {
-    (*c->plain_cleanup_cb)(c->data);
+    if (c->plain_cleanup_cb) {
+      (*c->plain_cleanup_cb)(c->data);
+    }
+
     c = c->next;
   }
 }
-
-/* NOTE: these cleanup routines are currently unused.
- * 2002-07-24
- */
-#if 0
-static void run_child_cleanups(cleanup_t *c) {
-  while (c) {
-    (*c->child_cleanup_cb)(c->data);
-    c = c ->next;
-  }
-}
-
-static void cleanup_pool_for_exec(pool *p) {
-  run_child_cleanups(p->cleanups);
-  p->cleanups = NULL;
-
-  for (p = p->sub_pools; p; p = p->sub_next)
-    cleanup_pool_for_exec(p);
-}
-
-void cleanup_for_exec(void) {
-  pr_alarms_block();
-  cleanup_pool_for_exec(permanent_pool);
-  pr_alarms_unblock();
-}
-#endif
-
-/*
- * Files and file descriptors
- */
-
-static void fd_cleanup_cb(void *fdv) {
-  close((int)fdv);
-}
-
-static void register_fd_cleanups(pool *p, int fd) {
-  register_cleanup(p, (void *)fd, fd_cleanup_cb, fd_cleanup_cb);
-}
-
-int popenf(pool *p, const char *name, int flags, int mode) {
-  int fd;
-
-  pr_alarms_block();
-  if ((fd = open(name, flags, mode)) >= 0)
-    register_fd_cleanups(p, fd);
-  pr_alarms_unblock();
-  return fd;
-}
-
-int pclosef(pool *p, int fd) {
-  int res;
-
-  pr_alarms_block();
-  res = close(fd);
-  unregister_cleanup(p, (void *)fd, fd_cleanup_cb);
-  pr_alarms_unblock();
-  return res;
-}
-
-/* Sep. plain and child cleanups for FILE *, since fclose() flushes
- * the stream
- */
-
-static void file_cleanup_cb(void *fpv) {
-  fclose((FILE *)fpv);
-}
-
-static void file_child_cleanup_cb(void *fpv) {
-  close(fileno((FILE *) fpv));
-}
-
-void register_file_cleanups(pool *p, FILE *fp) {
-  register_cleanup(p, (void *)fp, file_cleanup_cb, file_child_cleanup_cb);
-}
-
-FILE *pfopen(pool *p, const char *name, const char *mode) {
-  FILE *fd = NULL;
-  int base_flag, desc;
-
-  pr_alarms_block();
-
-  if (*mode == 'a') {
-    base_flag = (*(mode+1) == '+') ? O_RDWR : O_WRONLY;
-    desc = open(name, base_flag|O_APPEND|O_CREAT,
-       S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
-    if (desc >= 0)
-      fd = fdopen(desc, mode);
-
-  } else
-    fd = fopen(name, mode);
-
-  if (fd)
-    register_file_cleanups(p, fd);
-
-  pr_alarms_unblock();
-  return fd;
-}
-
-FILE *pfdopen(pool *p, int fd, const char *mode) {
-  FILE *f;
-
-  pr_alarms_block();
-  if ((f = fdopen(fd, mode)) != NULL)
-    register_file_cleanups(p, f);
-
-  pr_alarms_unblock();
-  return f;
-}
-
-int pfclose(pool *p, FILE *fd) {
-  int res;
-
-  pr_alarms_block();
-  res = fclose(fd);
-  unregister_cleanup(p, (void *) fd, file_cleanup_cb);
-  pr_alarms_unblock();
-  return res;
-}
